//+------------------------------------------------------------------+
//|                                              ExprBotPSOEmbed.mq5 |
//|                                    Copyright (c) 2020, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| Parallel Particle Swarm Optimization                             |
//|                           https://www.mql5.com/ru/articles/8321/ |
//|                                                                  |
//| Demo for simplified process of embedding PPSO in an EA using     |
//| ParticleSwarmEmbed.mqh                                           |
//| (Settings, worker Functors, and event handlers included)         |
//+------------------------------------------------------------------+
#property copyright "Copyright (c) 2020, Marketeer"
#property link      "https://www.mql5.com/en/users/marketeer"
#property version   "1.0"

#property tester_no_cache

// this must be defined for parallel PSO to run
// this can not be defined in the header, because it should get EA's mq5 file
// this can not be a variable, because it's used in the #property tester_file directive
#define PPSO_SHARED_SETTINGS __FILE__ + ".csv"

// property takes effect only when specified in the main mq5 file
#property tester_file PPSO_SHARED_SETTINGS


#define INDICATOR_FUNCTORS
#define PSO_LOG_VERBOSE

sinput string GroupTradingStrategy = ""; // >>  T R A D I N G   S T R A T E G Y
input string SignalBuy = "M_EMA_OPEN_9_6_27(Fast,Bar)/M_EMA_OPEN_9_6_27(Slow,Bar) > 1 + T";
input string SignalSell = "M_EMA_OPEN_9_6_27(Fast,Bar)/M_EMA_OPEN_9_6_27(Slow,Bar) < 1 - T";
input int Fast = 9;
input int Slow = 21;
input double T = 0.0; // Threshold
input double Lot = 0.1;


sinput string GroupVirtualLib = ""; // >>  V I R T U A L
sinput bool VirtualTester = false;
sinput ENUM_STATISTICS Estimator = STAT_PROFIT;
sinput string Formula = "";


sinput string GroupPSO = ""; // >>  P S O
sinput bool PSO_Enable = false;


#include <ParticleSwarmEmbed.mqh>
#include <fxsaber/MT4Orders.mqh>
#include <fxsaber/Virtual/Virtual.mqh>
#include <ExpresSParserS/v1.1/ExpressionCompiler.mqh>
#include <ExpresSParserS/v1.1/TesterStats.mqh>


VariableTable vts, vtf; // trade signals, target function
ExpressionCompiler ecb(vts), ecs(vts), ece(vtf);
Promise *p1, *p2, *p3;
MqlTick _ticks[];


class CustomFunctor: public BaseFunctor
{
  public:
    CustomFunctor(const int n): BaseFunctor(n) {}
    virtual double calculate(const double &vec[])
    {
      // assert ArraySize(vec) == params
      string msg = "";
      for(int i = 0; i < params; i++)
      {
        vts.set(names[i], vec[i]);
        msg += names[i] + ":" + (string)(float)vec[i] + ", ";
      }
      
      VIRTUAL::Tester(_ticks, OnTick, TesterStatistics(STAT_INITIAL_DEPOSIT));
      VIRTUAL::ResetTickets();
      const double r = p3 ? p3.with(EmbedTesterStats).resolve() : TesterStatistics(Estimator);
      msg += (string)(float)r;
      Print(msg);
      VIRTUAL::Delete();
      return r;
    }
};


int OnInit()
{
  if(PSO_Enable)
  {
    if(!VirtualTester)
    {
      Print("Internal optimization (PSO_Enable) mode is ignored: it's available with VirtualTester only");
    }
    else
    {
      const int status = PPSO_EventHandlers::OnInit(PPSO_SHARED_SETTINGS);
      if(status != INIT_SUCCEEDED) return status;
    }
  }

  vts.set("Fast", Fast);
  vts.set("Slow", Slow);
  vts.set("T", T);
  vts.set("Bar", 0);

  p1 = ecb.evaluate(SignalBuy, true);
  if(!ecb.success())
  {
    Print("Syntax error in Buy signal:");
    p1.print();
    return INIT_FAILED;
  }

  p2 = ecs.evaluate(SignalSell, true);
  if(!ecs.success())
  {
    Print("Syntax error in Sell signal:");
    p2.print();
    return INIT_FAILED;
  }

  if(StringLen(Formula) > 0)
  {
    EmbedTesterStats(vtf);
    p3 = ece.evaluate(Formula);
    if(!ece.success())
    {
      Print("Syntax error in Formula:");
      p3.print();
      return INIT_FAILED;
    }
  }
  else
  {
    p3 = NULL;
  }
  
  return INIT_SUCCEEDED;
}


bool IsNewBar(datetime now)
{
  static datetime lastBar = 0;
  
  now = now / PeriodSeconds() * PeriodSeconds();

  if(lastBar != now)
  {
    lastBar = now;
    return true;
  }
  return false;
}


#define _Ask SymbolInfoDouble(_Symbol, SYMBOL_ASK)
#define _Bid SymbolInfoDouble(_Symbol, SYMBOL_BID)

bool OnTesterCalled = false;

void OnTick()
{
  if(VirtualTester && !OnTesterCalled)
  {
    MqlTick _tick;
    SymbolInfoTick(_Symbol, _tick);
    const int n = ArraySize(_ticks);
    ArrayResize(_ticks, n + 1, n / 2);
    _ticks[n] = _tick;
    
    return; // skip all time scope and collect ticks
  }

  if(!IsNewBar(TimeCurrent())) return;
  
  const int bar = iBarShift(_Symbol, PERIOD_CURRENT, TimeCurrent());

  bool buy = p1.with("Bar", bar).resolve();
  bool sell = p2.with("Bar", bar).resolve();
  
  if(buy && sell)
  {
    buy = false;
    sell = false;
  }
  
  if(buy)
  {
    OrdersCloseAll(_Symbol, OP_SELL);
    if(OrdersTotalByType(_Symbol, OP_BUY) == 0)
    {
      OrderSend(_Symbol, OP_BUY, Lot, _Ask, 100, 0, 0);
    }
  }
  else if(sell)
  {
    OrdersCloseAll(_Symbol, OP_BUY);
    if(OrdersTotalByType(_Symbol, OP_SELL) == 0)
    {
      OrderSend(_Symbol, OP_SELL, Lot, _Bid, 100, 0, 0);
    }
  }
  else
  {
    OrdersCloseAll();
  }
}


void OrdersCloseAll(const string symbol = NULL, const int type = -1) // OP_BUY or OP_SELL
{
  for(int i = OrdersTotal() - 1; i >= 0; i--)
  {
    if(OrderSelect(i, SELECT_BY_POS))
    {
      if(OrderType() <= OP_SELL
      && (type == -1 || OrderType() == type)
      && (symbol == NULL || symbol == OrderSymbol()))
      {
        OrderClose(OrderTicket(), OrderLots(), OrderClosePrice(), 100);
      }
    }
  }
}

int OrdersTotalByType(const string symbol = NULL, const int type = -1) // OP_BUY or OP_SELL
{
  int count = 0;
  for(int i = OrdersTotal() - 1; i >= 0; i--)
  {
    if(OrderSelect(i, SELECT_BY_POS))
    {
      if(OrderType() <= OP_SELL
      && (type == -1 || OrderType() == type)
      && (symbol == NULL || symbol == OrderSymbol()))
      {
        count++;
      }
    }
  }
  return count;
}


void OnTesterInit()
{
  if(VirtualTester)
  {
    if(!PSO_Enable)
    {
      Print("Internal optimization in brute force mode makes no sense within built-in optimization");
      Print("Enable PSO mode (PSO_Enable=true)");
      ExpertRemove();
      return;
    }
    PPSO_EventHandlers::OnTesterInit(PPSO_SHARED_SETTINGS);
  }
}

void OnTesterPass()
{
  PPSO_EventHandlers::OnTesterPass();
}

void OnTesterDeinit()
{
  PPSO_EventHandlers::OnTesterDeinit();
}

double OnTester()
{
  if(VirtualTester)
  {
    OnTesterCalled = true;

    // MQL API implies some limitations for this function, so ticks are collected in OnTick
    // const int size = CopyTicksRange(_Symbol, _ticks, COPY_TICKS_ALL);

    const int size = ArraySize(_ticks);
    PrintFormat("Ticks size=%d error=%d", size, GetLastError());
    if(size <= 0) return 0;
    
    if(PSO_Enable)
    {
      return PPSO_EventHandlers::OnTester<CustomFunctor>();
    }
    else
    {
      Print("Fallback to single test: PSO ", PSO_Enable);
      VIRTUAL::Tester(_ticks, OnTick, TesterStatistics(STAT_INITIAL_DEPOSIT));
      Print(VIRTUAL::ToString(INT_MAX));
      Print("Trades: ", VIRTUAL::VirtualOrdersHistoryTotal());
      return p3 ? p3.with(EmbedTesterStats).resolve() : TesterStatistics(Estimator);
    }
    return 0;
  }
  return p3 ? p3.with(EmbedTesterStats).resolve() : TesterStatistics(Estimator);
}

void OnDeinit(const int)
{
}